// DlgProperties.cpp


#include "stdafx.h"
#include "SimuRoute.h"
#include "DlgProperties.h"


static void DDX_MyText(CDataExchange* pDX, int nIDC, int& value, int min, int max);


CDlgProperties e_dlgProp;


const UINT WM_MYUPDATE = WM_APP + 1;


BEGIN_MESSAGE_MAP(CDlgProperties, CDialog)
	ON_WM_CLOSE()
	ON_MESSAGE(WM_MYUPDATE, OnUpdate)
END_MESSAGE_MAP()


// Data
void CDlgProperties::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	
#define DDX_MYTEXT(name, min,max, def, id) \
		DDX_MyText(pDX, id, m_nd.name, min,max);
	NODEDATA_ENUM(DDX_MYTEXT)
#undef COMPARE
}


// Data exchange with "empty value" (empty value = min - 1)
void DDX_MyText(CDataExchange* pDX, int nIDC, int& value, int min, int max)
{
	if (min==0 && max==1) {
		// Checkbox
		
		DDX_Check(pDX, nIDC, value);
		return;
	}
	
	const HWND hwndEdit = pDX->PrepareEditCtrl(nIDC);
	
	if (pDX->m_bSaveAndValidate) {
		// Get value from the dialog box
		
		if (::GetWindowTextLength(hwndEdit)==0) {
			value = min - 1;
			return;
		}
		
	}else{
		// Put value in the dialog box
		
		if (value==min-1) {
			::SetWindowText(hwndEdit, NULL);
			return;
		}
		
		if (value < min || value > max)
			value = min;
	}
	
	// Normal
	DDX_Text     (pDX, nIDC, value);
	DDV_MinMaxInt(pDX,       value, min, max);
}


// Initialization
BOOL CDlgProperties::OnInitDialog()
{
	Update(NULL);
	
	CDialog::OnInitDialog();
	
	// Load placement in the settings
	WINDOWPLACEMENT *pwp;
	UINT len;
	if (app.GetProfileBinary(pszMain, pszPropPos, (BYTE**)&pwp, &len)) {
		if (len==sizeof(*pwp)) {
			SetWindowPos(NULL,
				pwp->rcNormalPosition.left,pwp->rcNormalPosition.top,
				0,0,
				SWP_NOZORDER | SWP_NOSIZE);
			ShowWindow(pwp->showCmd);
		}
	}else
		ShowWindow(SW_SHOW);
	delete [] (BYTE*)pwp;
	
	return TRUE;
}


// Closing -> hide window
void CDlgProperties::OnClose()
{
	ShowWindow(SW_HIDE);
}


// Save placement in the settings
void CDlgProperties::SavePlacement() const
{
	WINDOWPLACEMENT wp;
	e_dlgProp.GetWindowPlacement(&wp);
	if (!IsWindowVisible())
		wp.showCmd = SW_HIDE;
	app.WriteProfileBinary(pszMain, pszPropPos, (BYTE*)&wp, sizeof(wp));
}


// Update the dialog box with a document selection
void CDlgProperties::Update(const CDoc* pDoc)
{
	m_nd = app.m_ndDefault;
	
	bool fFirst = true;
	if (pDoc) {
		POSITION pos = pDoc->m_lstNode.GetHeadPosition();
		while (pos) {
			const CNode *const pNode = pDoc->m_lstNode.GetNext(pos);
			if (!pNode->m_sel)
				continue;
			const NODEDATA &nd = pNode->m_nd;
			
			if (fFirst) {
				fFirst = false;
				m_nd = pNode->m_nd;
				continue;
			}
			
			#define COMPARE(name, min,max, def, id) \
				if (m_nd.name != nd.name) \
					m_nd.name = min-1;
			NODEDATA_ENUM(COMPARE)
			#undef COMPARE
		}
	}
	
	
	UpdateData(FALSE);
	
	
	SetWindowText(CString((fFirst)
		? MAKEINTRESOURCE(IDS_PROPERTIES_DEFAULT)
		: MAKEINTRESOURCE(IDS_PROPERTIES)));
}


// Command message
BOOL CDlgProperties::OnCommand(WPARAM wParam, LPARAM lParam)
{
	if (HIWORD(wParam)==EN_KILLFOCUS || HIWORD(wParam)==BN_CLICKED)
		PostMessage(WM_MYUPDATE);
	
	return CDialog::OnCommand(wParam, lParam);
}


// Update
LRESULT CDlgProperties::OnUpdate(WPARAM, LPARAM)
{
	if (!UpdateData())
		// Error
		return 0;
	
	CFrameWnd *const pFrame = ((CFrameWnd*)app.m_pMainWnd)->GetActiveFrame();
	CDoc *const pDoc = (pFrame) ? (CDoc*)pFrame->GetActiveDocument() : NULL;
	bool fStored = false;
	if (pDoc) {
		POSITION pos = pDoc->m_lstNode.GetHeadPosition();
		while (pos) {
			CNode *const pNode = pDoc->m_lstNode.GetNext(pos);
			if (pNode->m_sel) {
				NODEDATA &nd = pNode->m_nd;
				
				#define SAVE(name, min,max, def, id) \
					if (m_nd.name != min-1) \
						nd.name = m_nd.name;
				NODEDATA_ENUM(SAVE);
				#undef COMPARE
				
				pDoc->SetModifiedFlag();
				fStored = true;
			}
		}
	}
	
	if (!fStored)
		app.m_ndDefault = m_nd;
	
	return 0;
}
