// GraphView.cpp


#include "stdafx.h"
#include "SimuRoute.h"

#include "Doc.h"
#include "GraphView.h"
#include "DlgProperties.h"
#include "DlgRebar.h"

#include "MainFrame.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#endif


const int cCellZoomStep  =  4;
const int mincCell       = cArrow + 2*cBorder;
const int maxcCell       = 60;

const UINT idTimerSimul  = 1;

const int iImageLight = 4;


static const COLORREF s_aclCar[] =
{
	RGB(  0,255,255), RGB(255,  0,  0), RGB(255,255,  0),
	RGB(255,  0,255), RGB(255,255,255), RGB(  0,255,  0),
	RGB(  0,  0,  0),
};
const int nbImageCar = nbArray(s_aclCar);
const int iImageSnail = nbImageCar-1;


IMPLEMENT_DYNCREATE(CGraphView, CScrollView)
BEGIN_MESSAGE_MAP(CGraphView, CScrollView)
	ON_COMMAND_RANGE          (ID_GRAPH_TURN_LEFT, ID_GRAPH_TURN_RIGHT, OnGraphTurn)
	ON_UPDATE_COMMAND_UI_RANGE(ID_GRAPH_TURN_LEFT, ID_GRAPH_TURN_RIGHT, OnGraphTurnUpdate)
	ON_COMMAND                (ID_VIEW_GRID, OnViewGrid)
	ON_UPDATE_COMMAND_UI      (ID_VIEW_GRID, OnUpdateViewGrid)
	ON_WM_ERASEBKGND()
	ON_WM_LBUTTONDOWN()
	ON_WM_RBUTTONDOWN()
	ON_WM_MOUSEMOVE()
	ON_WM_LBUTTONUP()
	ON_WM_SETCURSOR()
	ON_COMMAND_RANGE          (ID_VIEW_ZOOM_IN, ID_VIEW_ZOOM_OUT, OnViewZoom)
	ON_UPDATE_COMMAND_UI_RANGE(ID_VIEW_ZOOM_IN, ID_VIEW_ZOOM_OUT, OnUpdateViewZoom)
	ON_COMMAND          (ID_VIEW_ERRORS,    OnViewErrors)
	ON_UPDATE_COMMAND_UI(ID_VIEW_ERRORS,    OnUpdateViewErrors)
	ON_COMMAND          (ID_SIMUL_ACTIVATE, OnSimulActivate)
	ON_UPDATE_COMMAND_UI(ID_SIMUL_ACTIVATE, OnUpdateSimulActivate)
	ON_COMMAND          (ID_SIMUL_PAUSE,    OnSimulPause)
	ON_UPDATE_COMMAND_UI(ID_SIMUL_PAUSE,    OnUpdateSimulPause)
	ON_WM_TIMER()
	ON_COMMAND(ID_SIMUL_ADDSNAIL, OnSimulAddSnail)
	ON_UPDATE_COMMAND_UI(ID_SIMUL_ADDSNAIL, OnUpdateSimulAddSnail)
	ON_COMMAND          (ID_SIMUL_LOGGING,  OnSimulLogging)
	ON_UPDATE_COMMAND_UI(ID_SIMUL_LOGGING,  OnUpdateSimulLogging)
	ON_COMMAND          (ID_SIMUL_LOGFILE,  OnSimulLogFile)
	ON_COMMAND_RANGE(ID_PERSO_PLUS, ID_PERSO_MINUS, OnPerso)
END_MESSAGE_MAP()


UINT CGraphView::s_idTool = ID_GRAPH_TOOL_SELECTION;

const COLORREF CGraphView::s_acl[selCount][brushCount-1] =
{
	{ RGB(192,192,192), RGB(  0,  0,  0), RGB(128,128,128) },
	{ RGB(255,192,192), RGB(192,  0,  0), RGB(255,  0,  0) },
	{ RGB(192,192,255), RGB(  0,  0,192), RGB(  0,  0,255) },
	{ RGB( 64, 64,255), RGB(  0,  0,128), RGB(  0,  0,192) },
};
CBrush CGraphView::s_abrHatched[brushCount];

CBrush CGraphView::s_brError(RGB(255,0,0));

CImageList CGraphView::s_imlArrow;



// Constructor
CGraphView::CGraphView()
: m_size(1,1), m_aCellShadow(NULL),
  m_aUnit(NULL),
  m_apCar(NULL), m_apCarNew(NULL),
  m_aCar(NULL),
  m_apUnitValid(NULL)
{
	SetCellSize(32);
	m_fSimul      = false;
	m_fGrid       = true;
	m_fViewErrors = true;
	
	m_fSimulRun   = false;
	m_iSimulSpeed = 20;
#ifdef PERSO
	m_iSimulSkip  = perso_timeSkip;
#else
	m_iSimulSkip  = 1;
#endif
	
	m_fLogging    = false;
	
	m_nodeCreate.m_sel  = selShadowInvalid;
	m_nodeCreate.m_pt.x =
	m_nodeCreate.m_pt.y = -10;
	
	
	if (!(HBRUSH)s_abrHatched[0]) {
		// Create drawing objects
		
		for (int i=0; i<nbArray(s_abrHatched); i++)
			s_abrHatched[i].CreateHatchBrush(
				HS_DIAGCROSS, s_acl[i][brushBorder]);
		
		s_imlArrow.Create(IDB_ARROW, cArrow, 0, RGB(128,0,128));
	}
}

// Destructor
CGraphView::~CGraphView()
{
	delete [] m_aCellShadow;
	delete [] m_aUnit;
	delete [] m_apCar;
	delete [] m_apCarNew;
	delete [] m_aCar;
	delete [] m_apUnitValid;
}


// Before creation
BOOL CGraphView::PreCreateWindow(CREATESTRUCT& cs)
{
	return CScrollView::PreCreateWindow(cs);
}


// First initialization
void CGraphView::OnInitialUpdate()
{
	CScrollView::OnInitialUpdate();
	
	OnUpdateBoardSize();
}


// Update
void CGraphView::OnUpdate(CView*, LPARAM lHint, CObject* pHint)
{
	switch (lHint) {
		
		// Update board size
		case updateBoardSize:
			OnUpdateBoardSize();
			break;
		
		// Redraw a node
		case updateRedrawNode:
			{
				CRect rc;
				((CNode*)pHint)->GetRect(this, rc, false);
				rc -= GetScrollPosition();
				RedrawWindow(rc, NULL, RDW_INVALIDATE | RDW_ERASE);
			}
			break;
		
		// Redraw all
		case updateRedrawAll:
			RedrawWindow();
			break;
		
		// The current tool has changed
		case updateOnToolChange:
			m_nodeCreate.SetType((s_idTool>=idNodeFirst && s_idTool<=idNodeLast)
				? (NODETYPE)(s_idTool - idNodeFirst) : nodeInvalid);
			m_nodeCreate.m_pt.x =
			m_nodeCreate.m_pt.y = -10;
			RedrawWindow();
			break;
	}
}


// Erase background
BOOL CGraphView::OnEraseBkgnd(CDC* pDC)
{
	CRect rcClient;
	GetClientRect(rcClient);
	
	const HBRUSH hBrush = GetSysColorBrush(COLOR_WINDOW);
	::FillRect(*pDC, CRect(m_size.cx,0, rcClient.right,rcClient.bottom), hBrush);
	::FillRect(*pDC, CRect(0,m_size.cy, m_size.cx,     rcClient.bottom), hBrush);
	
	return TRUE;
}


// Draw
void CGraphView::OnDraw(CDC* pDC)
{
	if (m_size.cx==1)
		OnUpdateBoardSize();
	
	
	if (m_fSimul) {
		// Simulation
		
		
		const int iDrawArrowOffset = (m_cUnit - cArrow) / 2;
		const CSize sizeDrawArrowOffset(iDrawArrowOffset, iDrawArrowOffset);
		const int cLightOffset = iDrawArrowOffset;
		
		
		// Draw the roads
		m_draw.m_dc.BitBlt(0,0, m_size.cx,m_size.cy,
			&m_road.m_dc, 0,0, SRCCOPY);
		
		// Draw the lights
		POSITION pos = m_lstLight.GetHeadPosition();
		while (pos) {
			const CLight &light = m_lstLight.GetNext(pos);
			
			CPoint pt = GetUnitView(light.m_ptDraw) + sizeDrawArrowOffset;
			CSize offDir = aoffDir[light.m_dirDraw];
			pt += CSize(cLightOffset*offDir.cx, cLightOffset*offDir.cy);
			
			s_imlArrow.Draw(&m_draw.m_dc, iImageLight + light.m_fGreen,
				pt, ILD_TRANSPARENT);
		}
		
		
		// Draw the units
		for (int x=0; x<m_sizeUnit.cx; x++) {
			for (int y=0; y<m_sizeUnit.cy; y++) {
				const CAR *const pCar = GetCar(m_apCar, CPoint(x,y));
				if (pCar)
					m_draw.m_dc.FillSolidRect(CRect(
						GetUnitView(x,y) + sizeDrawArrowOffset,
						CSize(cCar,cCar)),
						s_aclCar[(pCar->maxSpeed==maxSpeedSnail)
							? iImageSnail
							: (pCar->id % iImageSnail)]);
			}
		}
		
	}else
		// Not the simulation
		DrawRoads(m_draw.m_dc, false);
	
	
	// Blit to the screen DC
	CRect rcUpdate;
	GetClientRect(rcUpdate);
	rcUpdate += GetScrollPosition();
	pDC->BitBlt(
		rcUpdate.left,rcUpdate.top,
		rcUpdate.Width(), rcUpdate.Height(),
		&m_draw.m_dc,
		rcUpdate.left,rcUpdate.top,
		SRCCOPY);
}


// Draw the roads in a given DC
void CGraphView::DrawRoads(CDC& dc, bool fSimul) const
{
	// Erase background
	::FillRect(dc, GetViewRect(), GetSysColorBrush(COLOR_WINDOW));
	
	// Draw grid
	if (m_fGrid) {
		for (int x=-1; x<m_size.cx; x+=m_cCell)
			for (int y=-1; y<m_size.cy; y+=m_cCell)
				dc.FillSolidRect(x,y, 2,2, RGB(192,192,192));
	}
	
	
	const CDoc *const pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	
	
	// Draw cells
	DrawCells(dc, pDoc->m_aCell, pDoc->GetBoardRect(), fSimul);
	
	
	// Draw errors
	if (!fSimul && m_fViewErrors) {
		const CLstError &lstError = pDoc->m_lstError;
		POSITION pos = lstError.GetHeadPosition();
		while (pos) {
			CRect rc = lstError.GetNext(pos);
			Convert(rc);
			rc.InflateRect(cBorderError+2, cBorderError+2);
			dc.FrameRect(rc, &s_brError);
		}
	}
	
	
	// Draw the "shadow node" for the creation tool
	if (!fSimul && s_idTool>=idNodeFirst && s_idTool<=idNodeLast &&
	    m_nodeCreate.m_type!=nodeInvalid)
		m_nodeCreate.Draw(this, dc, false);
	
	
	if (fSimul) {
		// Draw the static units (the arrows)
		
		const int iDrawArrowOffset = (m_cUnit - cArrow) / 2;
		const CSize sizeDrawArrowOffset(iDrawArrowOffset, iDrawArrowOffset);
		
		for (int x=0; x<m_sizeUnit.cx; x++) {
			for (int y=0; y<m_sizeUnit.cy; y++) {
				const UNIT &unit = GetUnit(x,y);
				if (!unit.maxSpeed)
					continue;
				
				// Find the direction with the maximum number of possible moves
				int maxnbMove  = 0;
				int imaxnbMove = 0;
				for (int i=0; i<8; i++) {
					const int nbMove = unit.aMoveDir[i].nbMove;
					if (maxnbMove < nbMove) {
						maxnbMove  = nbMove;
						imaxnbMove = i;
					}
				}
				
				if (maxnbMove==1)
					// Arrow
					s_imlArrow.Draw(&dc,
						unit.aMoveDir[imaxnbMove].aMove[0].dir,
						GetUnitView(x,y) + sizeDrawArrowOffset,
						ILD_TRANSPARENT);
			}
		}
	}
}


// Diagnostics
#ifdef _DEBUG
CDoc* CGraphView::GetDocument() const
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CDoc)));
	return (CDoc*)m_pDocument;
}
#endif


// Set zoom factor (cell size)
void CGraphView::SetCellSize(int cCell)
{
	m_cCell = cCell;
	m_cUnit = cCell / 2;
	
	ComputeBorderRect(m_arcCellBorder,      cBorder);
	ComputeBorderRect(m_arcCellBorderInt,   cBorder + cBorderInt);
	ComputeBorderRect(m_arcCellBorderLimit, cBorderInt);
	
	ComputeCornerRect(m_arcCellCorner,      cBorder);
	ComputeCornerRect(m_arcCellCornerInt,   cBorder + cBorderInt);
	
	if (m_pDocument) {
		
		// Update scroll size
		m_size = GetDocument()->m_size;
		m_size.cx *= m_cCell;
		m_size.cy *= m_cCell;
		
		// Create the offscreen bitmaps
		CDC *const pdc = GetDC();
		m_draw.Create(pdc, m_size);
		m_road.Create(pdc, m_size);
		ReleaseDC(pdc);
		
		SimulUpdate(true);
	}
	
	SetScrollSizes(MM_TEXT, m_size);
}



// Board size modified
void CGraphView::OnUpdateBoardSize()
{
	if (m_pDocument) {
		
		const CSize size = GetDocument()->m_size;
		m_sizeUnit.SetSize(size.cx*2, size.cy*2);
		
		delete [] m_aCellShadow;
		m_aCellShadow = new CELL[size.cx*size.cy];
		
		delete [] m_aUnit;
		m_aUnit = new UNIT[m_sizeUnit.cx*m_sizeUnit.cy];
		
		delete [] m_apCar;
		delete [] m_apCarNew;
		m_apCar    = new CAR*[m_sizeUnit.cx*m_sizeUnit.cy];
		m_apCarNew = new CAR*[m_sizeUnit.cx*m_sizeUnit.cy];
	}
	
	SetCellSize(m_cCell);
}


// Left button down
void CGraphView::OnLButtonDown(UINT nFlags, CPoint pt)
{
	if (m_fSimul && s_idTool!=ID_GRAPH_TOOL_HAND)
		return;
	
	CDoc *const pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	
	const CPoint ptNoScroll = pt;
	m_ptMouseOld = pt;
	pt += GetScrollPosition();
	const CPoint ptCell = GetToCell(pt);
	const CELL *const pCell = pDoc->GetSafeCell(ptCell);
	
	const bool fControl = ToBool(nFlags & MK_CONTROL);
	
	
	switch (s_idTool) {
		
		
		// Hand
		case ID_GRAPH_TOOL_HAND:
			SetCapture();
			break;
		
		
		// Selection
		case ID_GRAPH_TOOL_SELECTION:
			{
				// See if we have a drag or a click
				CPoint ptScreen = ptNoScroll;
				ClientToScreen(&ptScreen);
				const bool fDrag = ToBool(DragDetect(ptScreen));
				
				if (!fControl && !(fDrag && pCell && pCell->GetSel()))
					pDoc->OnEditUnselect();
				
				if (pCell && pCell->IsRoad() && pCell->pNode) {
					// Modify the aimed node selected state
					
					pCell->pNode->m_sel = (fControl)
						? (NODESEL)!pCell->pNode->m_sel
						: selYes;
					pDoc->UpdateAllViews(NULL, updateRedrawAll);
					pDoc->OnSelectionChanged();
				}
				
				if (!fDrag)
					// Single click: terminated now
					// NOT break because OnMouseMove() MUST be ignored
					return;
				
				// New drag/drop: move or selection
				m_fMove   = pCell && pCell->IsRoad();
				m_ptStart = pt;
				m_sizeMoveOld.SetSize(0,0);
				SetCapture();
			}
			break;
		
		
		// Other
		default:
			if (s_idTool>=idNodeFirst && s_idTool<=idNodeLast) {
				// Node creation
				
				if (s_idTool==ID_GRAPH_TOOL_NODE_LINE) {
					// Line: we need the other end of the line
					
					m_nodeCreate.m_pt = GetToCell(pt);
					m_nodeCreate.SetPoint2(m_nodeCreate.m_pt, pt);
					SetCapture();
					
				}else if (m_nodeCreate.m_sel==selShadow)
					// Other nodes: add now
					
					pDoc->AddNode(m_nodeCreate);
			}
			break;
	}
	
	OnMouseMove(nFlags, ptNoScroll);
}


// Right button down
void CGraphView::OnRButtonDown(UINT, CPoint)
{
	if (s_idTool==ID_GRAPH_TOOL_NODE_T)
		OnGraphTurn(ID_GRAPH_TURN_RIGHT);
}


// Mouse move
void CGraphView::OnMouseMove(UINT nFlags, CPoint pt)
{
	if (m_fSimul && s_idTool!=ID_GRAPH_TOOL_HAND)
		return;
	
	const CPoint ptNoScroll = pt;
	const CPoint ptScrollPos = GetScrollPosition();
	pt += ptScrollPos;
	
	const CPoint ptCell = GetToCell(pt);
	
	
	switch (s_idTool) {
		
		// Hand
		case ID_GRAPH_TOOL_HAND:
			if (nFlags & MK_LBUTTON)
				OnScrollBy(m_ptMouseOld - ptNoScroll);
			break;
		
		
		// Selection
		case ID_GRAPH_TOOL_SELECTION:
			if (nFlags & MK_LBUTTON) {
				
				if (m_fMove) {
					// Move
					
					CSize sizeMove = pt - m_ptStart;
					ToCell((POINT&)sizeMove);
					
					if (sizeMove!=m_sizeMoveOld) {
						// Try to move
						
						CSize sizeTrueMove = sizeMove - m_sizeMoveOld;
						
						CDoc *const pDoc = GetDocument();
						CLstNode &lstNode = pDoc->m_lstNode;
						POSITION pos;
						
						// Verify we can move
						pos = lstNode.GetHeadPosition();
						while (pos) {
							CNode *const pNode = lstNode.GetNext(pos);
							if (pNode->m_sel) {
								pNode->m_pt += sizeTrueMove;
								const bool fOK = pDoc->IsOverlapOK(*pNode, true);
								pNode->m_pt -= sizeTrueMove;
								
								if (!fOK)
									goto End;
							}
						}
						
						// OK, do the move now
						pos = lstNode.GetHeadPosition();
						while (pos) {
							CNode *const pNode = lstNode.GetNext(pos);
							if (pNode->m_sel)
								pNode->m_pt += sizeTrueMove;
						}
						
						pDoc->UpdateCells();
						m_sizeMoveOld = sizeMove;
						
	End:
						;
					}
					
					
				}else{
					// Select
					
					// Redraw the window (erase the old selection rect)
					RedrawWindow();
					
					// Draw selection rect
					CRect rc(m_ptStart, pt);
					rc.NormalizeRect();
					rc &= GetViewRect();
					rc -= ptScrollPos;
					CDC *const pdc = GetDC();
					pdc->DrawFocusRect(rc);
					ReleaseDC(pdc);
				}
			}
			break;
		
		
		// Other
		default:
			if (s_idTool>=idNodeFirst && s_idTool<=idNodeLast) {
				// Node creation
				
				const CSize sizeOrient(
					(pt.x % m_cCell) - m_cCell/2,
					(pt.y % m_cCell) - m_cCell/2);
				
				if (s_idTool==ID_GRAPH_TOOL_NODE_LINE) {
					if (!(nFlags & MK_LBUTTON))
						m_nodeCreate.m_pt = ptCell;
					m_nodeCreate.SetPoint2(ptCell, sizeOrient);
				}else{
					m_nodeCreate.m_pt = ptCell;
					
					if (m_nodeCreate.GetCellsSize()==CSize(0,0))
						m_nodeCreate.SetOrient(sizeOrient);
				}
				
				if (s_idTool==ID_GRAPH_TOOL_NODE_T || s_idTool==ID_GRAPH_TOOL_NODE_X) {
					m_nodeCreate.m_pt.x--;
					m_nodeCreate.m_pt.y--;
				}
				
				// See if it is valid to place the node here
				m_nodeCreate.m_sel = (GetDocument()->IsOverlapOK(m_nodeCreate))
					? selShadow : selShadowInvalid;
				
				RedrawWindow();
			}
			break;
	}
	
	m_ptMouseOld = ptNoScroll;
}


// Left button up
void CGraphView::OnLButtonUp(UINT nFlags, CPoint pt)
{
	if (m_fSimul && s_idTool!=ID_GRAPH_TOOL_HAND)
		return;
	
	ReleaseCapture();
	
	CDoc *const pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	
	const bool fControl = ToBool(nFlags & MK_CONTROL);
	
	switch (s_idTool) {
		
		
		// Selection
		case ID_GRAPH_TOOL_SELECTION:
			{
				if (m_fMove)
					break;
				
				if (!fControl)
					pDoc->OnEditUnselect();
				
				const CRect rcClip(CPoint(0,0), m_size);
				
				CRect rc(m_ptStart, pt);
				rc.NormalizeRect();
				
				if (rc.IsRectEmpty()) {
					// Single click
					
					if (rcClip.PtInRect(pt)) {
						CNode *const pNode = pDoc->GetCell(GetToCell(pt)).pNode;
						if (pNode) {
							pNode->m_sel = (fControl)
								? (NODESEL)!pNode->m_sel
								: selYes;
							pDoc->OnSelectionChanged();
						}
					}
					
				}else{
					// True rectangle
					
					rc &= rcClip;
					rc.left += m_cCell-1;
					rc.top  += m_cCell-1;
					ToCell(rc.TopLeft());
					ToCell(rc.BottomRight());
					
					for (int x=rc.left; x<=rc.right; x++) {
						for (int y=rc.top; y<=rc.bottom; y++) {
							if (!pDoc->IsPointInBoard(CPoint(x,y)))
								continue;
							
							CNode *const pNode = pDoc->GetCell(x,y).pNode;
							if (pNode)
								pNode->m_sel = selYes;
						}
					}
					
					pDoc->OnSelectionChanged();
				}
				
				// Erase selection rectangle
				RedrawWindow();
			}
			break;
		
		
		// Line node
		case ID_GRAPH_TOOL_NODE_LINE:
			pDoc->AddNode(m_nodeCreate);
			break;
	}
}


// Mouse cursor
BOOL CGraphView::OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT uMsg)
{
	HCURSOR hCursor = NULL;
	
	if (nHitTest==HTCLIENT) {
		
		CDoc *const pDoc = GetDocument();
		
		CPoint pt;
		GetCursorPos(&pt);
		ScreenToClient(&pt);
		pt += GetScrollPosition();
		
		const CPoint ptCell = GetToCell(pt);
		const CELL *pCell = pDoc->GetSafeCell(ptCell);
		
		
		switch (s_idTool) {
			
			// Hand
			// Left button -> hand cursor
			case ID_GRAPH_TOOL_HAND:
				if (HIWORD(GetAsyncKeyState((GetSystemMetrics(SM_SWAPBUTTON))
				    ? VK_RBUTTON : VK_LBUTTON)))
					hCursor = app.LoadCursor(IDC_HAND);
				break;
			
			// Selection
			// Over a selected element -> move cursor
			case ID_GRAPH_TOOL_SELECTION:
				if (!m_fSimul && pCell && pCell->IsRoad() && pCell->GetSel())
					hCursor = app.LoadStandardCursor(IDC_SIZEALL);
				break;
		}
	}
	
	if (!hCursor)
		return CScrollView::OnSetCursor(pWnd, nHitTest, uMsg);
	
	SetCursor(hCursor);
	return TRUE;
}


// Grid
void CGraphView::OnViewGrid()
{
	m_fGrid = !m_fGrid;
	RedrawWindow();
}

// Grid - update
void CGraphView::OnUpdateViewGrid(CCmdUI* pCmdUI)
{
	pCmdUI->SetCheck(m_fGrid);
}


// Turn left/right
void CGraphView::OnGraphTurn(UINT id)
{
	const int iOffset = (id==ID_GRAPH_TURN_LEFT) ? 3 : 1;
	
	CDoc *const pDoc = GetDocument();
	CLstNode &lstNode = pDoc->m_lstNode;
	POSITION pos = lstNode.GetHeadPosition();
	while (pos) {
		CNode *const pNode = lstNode.GetNext(pos);
		if (pNode->m_sel)
			pNode->Turn(pDoc, iOffset);
	}
	
	if (m_nodeCreate.m_type==nodeT)
		m_nodeCreate.Turn(pDoc, iOffset, false);
	
	pDoc->UpdateCells(true);
}

// Turn left/right - update
void CGraphView::OnGraphTurnUpdate(CCmdUI* pCmdUI)
{
	CLstNode &lstNode = GetDocument()->m_lstNode;
	
	bool fNodeSel = (s_idTool==ID_GRAPH_TOOL_NODE_T);
	
	POSITION pos = lstNode.GetHeadPosition();
	while (pos && !fNodeSel) {
		if (lstNode.GetNext(pos)->m_sel)
			fNodeSel = true;
	}
	
	pCmdUI->Enable(fNodeSel);
}


// Zoom
void CGraphView::OnViewZoom(UINT id)
{
	SetCellSize(m_cCell + ((id==ID_VIEW_ZOOM_IN) ? +cCellZoomStep : -cCellZoomStep));
	RedrawWindow();
}

// Zoom - Update
void CGraphView::OnUpdateViewZoom(CCmdUI* pCmdUI)
{
	pCmdUI->Enable((pCmdUI->m_nID==ID_VIEW_ZOOM_IN)
		? m_cCell+cCellZoomStep <= maxcCell
		: m_cCell-cCellZoomStep >= mincCell);
}


// Scale an "unitary" rectangle (example: {0,1, 1,1}) to a border rectangle
void CGraphView::ComputeBorderRect(CRect arc[], int cBorder)
{
	const int cBorderLength = m_cCell - cBorder;
	
	for (int iOrient=0; iOrient<4; iOrient++) {
		const CRect &rcFrom = arcBorder[iOrient];
		
		CRect &rcDest = arc[iOrient];
		rcDest.left   = rcFrom.left   * cBorderLength;
		rcDest.top    = rcFrom.top    * cBorderLength;
		rcDest.right  = rcFrom.right  * cBorderLength + cBorder;
		rcDest.bottom = rcFrom.bottom * cBorderLength + cBorder;
	}
}

// Same as ComputeBorderRect() but for corner rectangles
void CGraphView::ComputeCornerRect(CRect arc[], int cBorder)
{
	const int cBorderLength = m_cCell - cBorder;
	
	for (int iOrient=0; iOrient<4; iOrient++) {
		CPoint pt = aoffDiagDir[iOrient];
		pt.x *= cBorderLength;
		pt.y *= cBorderLength;
		arc[iOrient] = CRect(pt, CSize(cBorder,cBorder));
	}
}


// Draw a cell at a given position
void CGraphView::DrawCell(CDC& dc, int x, int y, const CELL& cell, bool fSimul) const
{
	if (!cell.pNode)
		return;
	
	const CPoint pt(x,y);
	const CPoint ptPixel = GetConvert(pt);
	
	const NODESEL sel = (fSimul) ? selNo : cell.pNode->m_sel;
	
	const int iWall  = cell.iWall;
	const int iLimit = iWall / maskLimitFirst;
	
	// Draw the road
	if (iWall & maskRoad) {
		ASSERT(cell.pNode);
		CRect rc(ptPixel, CSize(m_cCell,m_cCell));
		
		if (cell.pNode->m_type==nodeEnd) {
			// Hatched
			
			dc.SetBkColor(s_acl[sel][CGraphView::brushRoad]);
			dc.FillRect(rc, &s_abrHatched[sel]);
			
		}else
			// Solid color
			dc.FillSolidRect(rc, s_acl[sel][CGraphView::brushRoad]);
	}
	
	// Draw the interior borders...
	for (int iOrient=0, iCurWall=1; iOrient<4; iOrient++, iCurWall*=2)
		if (iWall & iCurWall)
			// Border
			dc.FillSolidRect(m_arcCellBorderInt[iOrient] + ptPixel,
				s_acl[sel][CGraphView::brushBorderInt]);
			
		else if (iLimit & iCurWall)
			// Limit
			dc.FillSolidRect(m_arcCellBorderLimit[iOrient] + ptPixel,
				s_acl[sel][CGraphView::brushBorderInt]);
	
	// ... And then the borders
	for (int iOrient=0, iCurWall=1; iOrient<4; iOrient++, iCurWall*=2)
		if (iWall & iCurWall)
			// Border
			dc.FillSolidRect(m_arcCellBorder[iOrient] + ptPixel,
				s_acl[sel][CGraphView::brushBorder]);
	
	
	// Draw the corners
	for (int iOrient=0, iCurPoint=maskPointFirst; iOrient<4; iOrient++, iCurPoint*=2)
		if (iWall & iCurPoint) {
			// Interior border, then normal border
			
			dc.FillSolidRect(m_arcCellCornerInt[iOrient] + ptPixel,
				s_acl[sel][CGraphView::brushBorderInt]);
			dc.FillSolidRect(m_arcCellCorner[iOrient] + ptPixel,
				s_acl[sel][CGraphView::brushBorder]);
		}
}


// Draw all cells of an array
void CGraphView::DrawCells(CDC& dc, CELL aCell[], CRect rc, bool fSimul) const
{
	const int cx = GetDocument()->m_size.cx;
	
	for (; rc.top<rc.bottom; rc.top++)
		for (int x=rc.left; x<rc.right; x++)
			DrawCell(dc, x,rc.top, aCell[x + rc.top*cx], fSimul);
}


// View errors
void CGraphView::OnViewErrors()
{
	m_fViewErrors = !m_fViewErrors;
	
	RedrawWindow();
}

// View errors - Update
void CGraphView::OnUpdateViewErrors(CCmdUI* pCmdUI)
{
	pCmdUI->SetCheck(m_fViewErrors);
}


// View activation
void CGraphView::OnActivateView(BOOL bActivate, CView* pActivateView, CView* pDeactiveView)
{
	e_dlgProp.Update((bActivate) ? GetDocument() : NULL);
	e_dlgRebar.SetSpeed((bActivate) ? m_iSimulSpeed : -1);
	e_dlgRebar.SetSkip ((bActivate) ? m_iSimulSkip  : -1);
	
	CScrollView::OnActivateView(bActivate, pActivateView, pDeactiveView);
}


// Simulation mode
void CGraphView::OnSimulActivate()
{
	if (m_fSimulRun)
		OnSimulPause();
	
	else{
		CDoc *const pDoc = GetDocument();
		
		pDoc->FindErrors();
#ifndef _DEBUG
		if (pDoc->m_lstError.GetCount() > 0) {
			AfxMessageBox(ERR_FINDERRORS);
			return;
		}
#endif
	}
	
	m_fSimul = !m_fSimul;
	
	if (m_fSimul) {
		
#ifndef PERSO
		// Empty the log file
		if (IsLogFileValid()) {
			m_fileLogging.SeekToBegin();
			SetEndOfFile(m_fileLogging);
		}
#endif
		
		SimulUpdate(false);
	}
	
	RedrawWindow();
	
	if (m_fSimul)
		OnSimulPause();
}

// Simulation mode - Update
void CGraphView::OnUpdateSimulActivate(CCmdUI* pCmdUI)
{
	pCmdUI->SetCheck(m_fSimul);
}



//*** Offscreen bitmap

// Create the bitmap
void CGraphView::COffscreen::Create(CDC* pDC, CSize size)
{
	Delete();
	
	// Create the DC and the bitmap
	m_dc.CreateCompatibleDC(pDC);
	m_bmp.CreateCompatibleBitmap(pDC, size.cx,size.cy);
	m_dc.SelectObject(&m_bmp);
}

// Delete the objets
void CGraphView::COffscreen::Delete()
{
	m_bmp.DeleteObject();
	m_dc .DeleteDC();
}


// Pause simulation
void CGraphView::OnSimulPause()
{
	if (m_fSimulRun)
		KillTimer(idTimerSimul);
	
	m_fSimulRun = !m_fSimulRun;
	
	if (m_fSimulRun)
		SetTimer(idTimerSimul, m_iSimulSpeed, NULL);
}

// Pause simulation
void CGraphView::OnUpdateSimulPause(CCmdUI* pCmdUI)
{
	pCmdUI->Enable(m_fSimul);
	pCmdUI->SetCheck(!m_fSimulRun);
}

// Update the timer
void CGraphView::SimulUpdateTimer(int iSimulSpeed)
{
	m_iSimulSpeed = iSimulSpeed;
	if (m_fSimulRun) {
		KillTimer(idTimerSimul);
		SetTimer(idTimerSimul, m_iSimulSpeed, NULL);
	}
}


// Timer event
void CGraphView::OnTimer(UINT nIDEvent)
{
	if (nIDEvent!=idTimerSimul) {
		CScrollView::OnTimer(nIDEvent);
		return;
	}
	
	// One computing at once
	static bool s_fCalcul = false;
	if (s_fCalcul)
		return;
	s_fCalcul = true;
	
	
	// Compute the states and the statistics
	const STATISTICS statsOld = m_stats;
	int nbAvgCar = 0;
	for (int i=0; i<m_iSimulSkip; i++) {
		nbAvgCar += m_nbCar;
		SimulTick();
	}
	
	STATISTICS statsDelta = m_stats;
	statsDelta.time           -= statsOld.time;
	statsDelta.nbCarCreated   -= statsOld.nbCarCreated;
	statsDelta.nbCarDestroyed -= statsOld.nbCarDestroyed;
	statsDelta.nbCarRefused   -= statsOld.nbCarRefused;
	statsDelta.timeCarLive    -= statsOld.timeCarLive;
	statsDelta.nbSnailCreated -= statsOld.nbSnailCreated;
	statsDelta.nbSnail        -= statsOld.nbSnail;
	nbAvgCar /= statsDelta.time;
	
	m_stats.avgCarCreated = (float)statsDelta.nbCarCreated / (float)statsDelta.time;
	m_stats.avgCarRefused = (float)statsDelta.nbCarRefused / (float)statsDelta.time;
	m_stats.avgCarLive    = (statsDelta.nbCarDestroyed)
		? (float)statsDelta.timeCarLive  / (float)statsDelta.nbCarDestroyed
		: 0;
	
	if (m_fLogging) {
		
		// Write the log line
		CString sLine;
#ifdef PERSO
		NODEDATA &ndStart = GetDocument()->GetCell(perso_ptStart).pNode->m_nd;
		NODEDATA &ndX     = GetDocument()->GetCell(perso_ptX1).pNode->m_nd;
		sLine.Format(_T("%d;%d;%d;%d;%d;%1.2f;%1.2f;%1.1f\r\n"),
			ndStart.m_probaNewCar,
			ndX.m_maxSpeed,
			ndX.m_timeG,
			ndX.m_timeR,
			ndX.m_delaySynchro,
			m_stats.avgCarCreated,
			m_stats.avgCarRefused,
			m_stats.avgCarLive);
#else
		sLine.Format(_T("%d;%d;%d;%d;%d;%d;%d;%1.1f;%1.2f\r\n"),
			statsOld.time,
			m_stats.nbCarCreated,
			nbAvgCar,
			statsDelta.nbCarCreated,
			statsDelta.nbCarDestroyed,
			statsDelta.nbSnail,
			statsDelta.nbSnailCreated,
			m_stats.avgCarLive,
			m_stats.avgCarCreated);
#endif
		sLine.Replace('.', ',');
		m_fileLogging.Write(sLine, sLine.GetLength());
	}
	
	s_fCalcul = false;
	
	// Redraw the window and update the GUI
	RedrawWindow();
	((CMainFrame*)app.m_pMainWnd)->m_wndStatusBar.OnIdleUpdateCmdUI(TRUE, 0);
	
#ifdef PERSO
	OnSimulActivate();
	
	CDoc *const pDoc = GetDocument();
	
	// Modify the nodes
	CNode *const pNode1 = pDoc->GetCell(perso_ptX1).pNode;
	CNode *const pNode2 = pDoc->GetCell(perso_ptX2).pNode;
	pNode1->m_nd.m_delaySynchro++;
	pNode2->m_nd.m_delaySynchro = pNode1->m_nd.m_delaySynchro * 2;
	pDoc->OnSelectionChanged();
	if (pNode1->m_nd.m_delaySynchro > perso_delaySynchroStop) {
		LogFileClose();
		return;
	}
	
	OnSimulActivate();
#endif
}


// Add a snail
void CGraphView::OnSimulAddSnail()
{
	m_nbWantSnail++;
}

// Add a snail - Update
void CGraphView::OnUpdateSimulAddSnail(CCmdUI* pCmdUI)
{
	pCmdUI->Enable(m_fSimul);
}


// Logging
void CGraphView::OnSimulLogging()
{
	m_fLogging = !m_fLogging;
}

// Logging - Update
void CGraphView::OnUpdateSimulLogging(CCmdUI* pCmdUI)
{
	pCmdUI->Enable(IsLogFileValid());
	pCmdUI->SetCheck(m_fLogging);
}


// Log file
void CGraphView::OnSimulLogFile()
{
	CString sPath;
	if (IsLogFileValid())
		sPath = m_fileLogging.GetFilePath();
	else{
		sPath = GetDocument()->GetPathName();
		PathRenameExtension(sPath.GetBuffer(MAX_PATH), _T(".csv"));
		sPath.ReleaseBuffer();
	}
	
	
	CFileDialog fd(FALSE, _T("csv"),
		sPath,
		OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT | OFN_PATHMUSTEXIST,
		CString(MAKEINTRESOURCE(IDS_FILTER_LOG)));
	if (fd.DoModal())
		LogFileOpen(fd.GetPathName());
}



// Close the log file
void CGraphView::LogFileClose()
{
	if (IsLogFileValid())
		m_fileLogging.Close();
	m_fLogging = false;
}

// Open the log file
void CGraphView::LogFileOpen(LPCTSTR pszFile)
{
	LogFileClose();
	
	m_fileLogging.Open(pszFile,
		CFile::modeCreate | CFile::modeWrite | CFile::shareDenyWrite);
	m_fLogging = IsLogFileValid();
}


// Customized command
void CGraphView::OnPerso(UINT)
{
#ifdef PERSO
	CDoc *const pDoc = GetDocument();
	
	// Initialize the nodes
	pDoc->GetCell(perso_ptX1).pNode->m_nd.m_delaySynchro =
	pDoc->GetCell(perso_ptX2).pNode->m_nd.m_delaySynchro = perso_delaySynchroInit;
	
	// Open log file
	LogFileOpen(_T("G:\\Programmes\\SimuRoute\\VagueVerteDcalage.csv"));
	
	// Start
	OnSimulActivate();
#endif
}
