// Node.h


#pragma once

class CGraphView;
class CDoc;
struct CELL;
typedef CList<CRect> CLstError;


enum NODESEL
{
	selNo,
	selYes,
	selShadow,
	selShadowInvalid,
	
	selCount
};


enum NODETYPE
{
	nodeEnd,      // Road end
	nodeCorner,   // Corner (single cell)
	nodeT,        // T
	nodeX,        // +
	nodeLine,     // Variable length line
	nodeInvalid
};


const int maxNodeLink = 4;


// Parameters given to macro:
// - member name
// - minimum value
// - maximum value
// - default value
#define NODEDATA_ENUM(macro) \
	macro(m_maxSpeed,     0,100,  3, IDCTXT_MAXSPEED) \
	macro(m_probaNewCar,  0,100, 10, IDCTXT_PROBANEWCAR) \
	macro(m_timeG,        0,100, 10, IDCTXT_TIMEG) \
	macro(m_timeR,        0,100, 10, IDCTXT_TIMER) \
	macro(m_delaySynchro, 0,100,  0, IDCTXT_DELAYSYNCHRO) \
	macro(m_fStats,       0,1,    1, IDCCHK_STATS) \
	for (int i=0; i<maxNodeLink; i++) \
		macro(m_aProba[i], 0,100, 10, IDCTXT_PROBA0 + i)


struct NODEDATA0
{
	int m_maxSpeed;      // Speed limit
	
	// nodeEnd
	int m_probaNewCar;   // In %
	
	// nodeX, nodeT
	int m_timeG;                 // Green light duration
	int m_timeR;                 // Red light duration
	int m_delaySynchro;          // First color change
	int m_aProba[maxNodeLink];   // Probability to go to each output
};


struct NODEDATA : public NODEDATA0
{
	// nodeEnd
	int  m_fStats;   // Take this node into account for statistics
};


class CNode : public CObject
{
	DECLARE_SERIAL(CNode)
	
public:  // Functions
	
	CNode();
	
	CNode& operator = (const CNode& node);
	
	void SetType(NODETYPE type);
	void Turn(CDoc* pDoc, int iOffset, bool fMustBeValid=true);
	
	inline bool IsHorz() const
	{
		return ToBool(m_iOrient & 1);
	}
	inline bool IsVert() const
	{
		return !(m_iOrient & 1);
	}
	
	
	void Serialize(CArchive& ar);
	
	CSize GetCellsSize(bool fStrict=true) const;
	CRect GetCellsRect(bool fStrict=true) const;
	
	void GetRect(const CGraphView* pView, CRect& rc, bool fStrict) const;
	void Draw   (const CGraphView* pView, CDC& dc, bool fSimul) const;
	
	void SetOrient(CSize sizeOrient);
	void SetPoint2(CPoint pt2, CSize sizeOrient);
	
	void UpdateCells(CELL aCell[], CSize size) const;
	void ModifyCell (CELL aCell[], CSize size, int x, int y, int iWall) const;
	
	void UpdateUnits(CGraphView* pView) const;
	
	
public:  // Data
	
	NODETYPE  m_type;
	CPoint    m_pt;        // Start point
	int       m_iOrient;   // Orientation: 0~3 -> 0~270
	int       m_len;   // 0 = one cell, 1 = two cells, etc.
	
	NODEDATA  m_nd;
	
	NODESEL   m_sel;
	
	
	int       m_nbLink;
};

typedef CTypedPtrList<CObList, CNode*> CLstNode;
